﻿unit Jpfix.Core;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Types, System.Rtti,
  Vcl.Controls;

type
  PEdPos = ^TEdPos;
  TEdPos = packed record
    Col: SmallInt;
    Line: Longint;
  end;

  PEkView = ^TEkView;
  TEkView = record
  private
  class var
    FOffsetCursor: Integer;
    FOffsetLoaded: Boolean;
    FOffsetWinLeftCol: Integer;

    class procedure Initialize; static;
    function GetCursor: TEdPos; inline;
    function GetWinLeftCol: SmallInt; inline;
  public
    property Cursor: TEdPos read GetCursor;
    property WinLeftCol: SmallInt read GetWinLeftCol;
  end;

  TEditView = class
  private
  class var
    FOffsetEkView: Integer;
    FOffsetLoaded: Boolean;

    class procedure Initialize; //class constructor Create; // W1025 Unsupported language feature: 'class constructor'
    function GetEkView: PEkView; inline;
  public
    property EkView: PEkView read GetEkView;
  end;

  TCustomEditControl = class(TCustomControl)
  private
  class var
    FOffsetEkView: Integer;
    FOffsetcxChar: Integer;
    FOffsetCurEditView: Integer;
    FOffsetLeftGutter: Integer;
    FOffsetHotLine: Integer;
    FOffsetsHScrollPos: Integer;
    FOffsetLoaded: Boolean;

    class procedure Initialize; //class constructor Create; // W1025 Unsupported language feature: 'class constructor'
    function GetEkView: PEkView; inline;
    function GetCharWidth: Integer; inline;
    function GetCurEditView: TEditView; inline;
    function GetLeftGutter: Integer; inline;
    function GetHotlinkLine: Integer; inline;
    function GetHScrollPos: Integer; inline;
  public
    function GetTextAtLine(LineNum: Integer): string;

    property EkView: PEkView read GetEkView;
    property CharWidth: Integer read GetCharWidth;
    property CurEditView: TEditView read GetCurEditView;
    property LeftGutter: Integer read GetLeftGutter;
    property HotlinkLine: Integer read GetHotlinkLine;
    property HScrollPos: Integer read GetHScrollPos;
  end;

function MyColFromX(Self: TCustomEditControl; X: Integer): Integer;
function MyGetCaretPixelX(Self: TCustomEditControl): Integer;

const
  sCoreideName = {$IFDEF VER250}'coreide180.bpl';{$ENDIF} // XE4
                 {$IFDEF VER260}'coreide190.bpl';{$ENDIF} // XE5
                 {$IFDEF VER270}'coreide200.bpl';{$ENDIF} // XE6
                 {$IFDEF VER280}'coreide210.bpl';{$ENDIF} // XE7
                 {$IFDEF VER290}'coreide220.bpl';{$ENDIF} // XE8
                 {$IFDEF VER300}'coreide230.bpl';{$ENDIF} // 10 Seattle
                 {$IFDEF VER310}'coreide240.bpl';{$ENDIF} // 10.1 Berlin
                 {$IFDEF VER320}'coreide250.bpl';{$ENDIF} // 10.2 Tokyo

implementation

function CalcDisplayWidth(S: string): Integer; external sCoreideName name '@Edprocs@CalcDisplayWidth$qqr20System@UnicodeString';

{ TEkView }

class procedure TEkView.Initialize;
var
  ctx: TRttiContext;
  typ: TRttiType;
  fld: TRttiField;
begin
  FOffsetLoaded := False;

  typ := ctx.FindType('EK.TEkView');
  if typ = nil then Exit;

  fld := typ.GetField('Cursor');
  if fld = nil then Exit;
  FOffsetCursor := fld.Offset;

  fld := typ.GetField('WinLeftCol');
  if fld = nil then Exit;
  FOffsetWinLeftCol := fld.Offset;

  FOffsetLoaded := True;
end;

function TEkView.GetCursor: TEdPos;
begin
  if not FOffsetLoaded then Exit(Default(TEdPos));
  Result := PEdPos(PByte(@Self) + FOffsetCursor)^;
end;

function TEkView.GetWinLeftCol: SmallInt;
begin
  if not FOffsetLoaded then Exit(0);
  Result := PSmallInt(PByte(@Self) + FOffsetWinLeftCol)^;
end;

{ TEditView }

class procedure TEditView.Initialize;
var
  ctx: TRttiContext;
  typ: TRttiType;
  fld: TRttiField;
begin
  FOffsetLoaded := False;

  typ := ctx.FindType('EditorBuffer.TEditView');
  if typ = nil then Exit;

  fld := typ.GetField('EkView');
  if fld = nil then Exit;
  FOffsetEkView := fld.Offset;

  FOffsetLoaded := True;
end;

function TEditView.GetEkView: PEkView;
begin
  if not FOffsetLoaded then Exit(nil);
  Result := PEkView(PPointer(PByte(Self) + FOffsetEkView)^);
end;

{ TCustomEditControl }

//class constructor TCustomEditControl.Create;
class procedure TCustomEditControl.Initialize;
var
  ctx: TRttiContext;
  typ: TRttiType;
  fld: TRttiField;
begin
  FOffsetLoaded := False;

  typ := ctx.FindType('EditorControl.TCustomEditControl');
  if typ = nil then Exit;

  fld := typ.GetField('FEkView');
  if fld = nil then Exit;
  FOffsetEkView := fld.Offset;

  fld := typ.GetField('cxChar');
  if fld = nil then Exit;
  FOffsetcxChar := fld.Offset;

  fld := typ.GetField('FCurEditView');
  if fld = nil then Exit;
  FOffsetCurEditView := fld.Offset;

  fld := typ.GetField('FLeftGutter');
  if fld = nil then Exit;
  FOffsetLeftGutter := fld.Offset;

  fld := typ.GetField('HotlinkLine');
  if fld = nil then Exit;
  FOffsetHotLine := fld.Offset;

  fld := typ.GetField('sHScrollPos');
  if fld = nil then Exit;
  FOffsetsHScrollPos := fld.Offset;

  FOffsetLoaded := True;
end;

function TCustomEditControl.GetTextAtLine; external sCoreideName name '@Editorcontrol@TCustomEditControl@GetTextAtLine$qqri';

function TCustomEditControl.GetEkView: PEkView;
begin
  if not FOffsetLoaded then Exit(nil);
  Result := PEkView(PPointer(PByte(Self) + FOffsetEkView)^);
end;

function TCustomEditControl.GetCharWidth: Integer;
begin
  if not FOffsetLoaded then Exit(0);
  Result := PInteger(PByte(Self) + FOffsetcxChar)^;
end;

function TCustomEditControl.GetCurEditView: TEditView;
begin
  if not FOffsetLoaded then Exit(nil);
  Result := TEditView(PPointer(PByte(Self) + FOffsetCurEditView)^);
end;

function TCustomEditControl.GetLeftGutter: Integer;
begin
  if not FOffsetLoaded then Exit(0);
  Result := PInteger(PByte(Self) + FOffsetLeftGutter)^;
end;

function TCustomEditControl.GetHotlinkLine: Integer;
begin
  if not FOffsetLoaded then Exit(0);
  Result := PInteger(PByte(Self) + FOffsetHotLine)^;
end;

function TCustomEditControl.GetHScrollPos: Integer;
begin
  if not FOffsetLoaded then Exit(0);
  Result := PInteger(PByte(Self) + FOffsetsHScrollPos)^;
end;

function MyColFromX(Self: TCustomEditControl; X: Integer): Integer;
var
  baseX, i, charWidth: Integer;
  s, ch: string;
  utf8: UTF8String;
begin
  baseX := X - Self.LeftGutter + Self.CharWidth div 2;
  Result := baseX div Self.CharWidth + Self.HScrollPos + 1;
  if Result > $1001 then
    Result := $1001
  else if Result < 1 then
    Result := 1;

  if Self.HotlinkLine > 0 then
  begin
    s := Self.GetTextAtLine(Self.HotlinkLine);
    if Length(s) = 0 then Exit;

    charWidth := Self.CharWidth;
    Result := 0;
    for i := 1 to High(Integer) do
    begin
      Dec(baseX, charWidth);
      if i <= Length(s) then
      begin
        case ByteType(s, i) of
          mbSingleByte: ch := Copy(s, i, 1);
          mbLeadByte: Continue;
          mbTrailByte: ch := Copy(s, i-1, 2);
        end;
        utf8 := UTF8String(ch);

        if (CalcDisplayWidth(ch) = 2) and (Length(utf8) = 3) then
          Inc(Result, 2)
        else
          Inc(Result, Length(utf8));
      end
      else
        Inc(Result);

      if baseX < 0 then
      begin
        Inc(Result, Self.HScrollPos);
        Break;
      end;
    end;
  end;
end;

function GetCorrectX(Self: TCustomEditControl; EdPos: TEdPos): Integer;
var
  col, missCol, textWidth: Integer;
  line: string;
begin
  line := Self.GetTextAtLine(EdPos.Line);
  col := 1;
  missCol := EdPos.Col;
  while (missCol > 0) and (col <= Length(line)) do
  begin
    case ByteType(line, col) of
      mbSingleByte:
        Dec(missCol, Length(UTF8String(line[col])));
      mbTrailByte:
        Dec(missCol, Length(UTF8String(Copy(line, col-1, 2))));
    end;
    Inc(col);
  end;
  textWidth := CalcDisplayWidth(Copy(line, 1, col-1));

  Result := Self.LeftGutter + (textWidth + missCol - Self.EkView^.WinLeftCol) * Self.CharWidth;
end;

function MyGetCaretPixelX(Self: TCustomEditControl): Integer;
begin
  if Self.CurEditView <> nil then
    Result := GetCorrectX(Self, Self.CurEditView.EkView^.Cursor)
  else
    Result := 0;
end;

initialization
  TEkView.Initialize;
  TEditView.Initialize;
  TCustomEditControl.Initialize;
end.
